/*
  samenamesicase.c - search case-insensitive same named files recursively

  Copyright (C) 2013 Taiji Yamada <taiji@aihara.co.jp>
  This program is distributed under the BSD 2-clause license.
  See http://opensource.org/licenses/BSD-2-Clause
*/
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>	/* offsetof() */
#include <string.h>
#include <sys/types.h>	/* dirent.h for the legacy */
#include <dirent.h>	/* struct dirent, etc */
#include <unistd.h>	/* pathconf(, _PC_NAME_MAX), etc */
#include <limits.h>	/* PATH_MAX */
#include <sys/stat.h>	/* S_ISLNK(), etc */
#include <errno.h>

int pathname_ignore_case_cmp(const void *a, const void *b)
{
  return strcasecmp(*(const char **)a, *(const char **)b);
}

int directory_tree(const char *pathname, int follow_symlinks)
{
  DIR *d;
  char *p, **v;
  struct dirent *entry, *result;
  size_t c = 0, i = 0, j, k, l;
  long name_max;

  if ((d = opendir(pathname))) {
    if (!((name_max = pathconf(pathname, _PC_NAME_MAX)) > 0))
      name_max = PATH_MAX;
    if (!(entry = malloc(offsetof(struct dirent, d_name) + name_max + 1))) {
      fprintf(stderr, "error: entry = malloc()\n");
      exit(1);
    }

    while (readdir_r(d, entry, &result) == 0 && result) {
      if (strcmp(entry->d_name, ".") == 0 ||
	  strcmp(entry->d_name, "..") == 0)
	continue;
      ++c;
    }
    rewinddir(d);
    k = c;

    v = malloc(sizeof(char **)*c);
    while (readdir_r(d, entry, &result) == 0 && result) {
      if (strcmp(entry->d_name, ".") == 0 ||
	  strcmp(entry->d_name, "..") == 0)
	continue;
      v[i] = malloc(strlen(entry->d_name) + 1);
      strcpy(v[i], entry->d_name);
      ++i;
    }
    rewinddir(d);

    qsort(v, c, sizeof(char *), pathname_ignore_case_cmp);
    for (i=0, j=1; j<c; ++i, ++j) {
      if (pathname_ignore_case_cmp(&v[i], &v[j]) == 0) {
	if (k != i) printf("%s/%s\n", pathname, v[i]);
	printf("%s/%s\n", pathname, v[(k=j)]);
      }
      else if (k != c) {
	printf("\n");
	k = c;
      }
    }
    if (k != c)
      printf("\n");
    for (i=0; i<c; ++i)
      free(v[i]);
    free(v);

    if (!(p = malloc((l=strlen(pathname)) + 1))) {
      fprintf(stderr, "error: p = malloc(strlen(pathname) + 1)\n");
      exit(1);
    }
    while (readdir_r(d, entry, &result) == 0 && result) {
      if (strcmp(entry->d_name, ".") == 0 ||
	  strcmp(entry->d_name, "..") == 0)
	continue;
      if (!(p = realloc(p, l + 1 + strlen(entry->d_name) + 1))) {
	fprintf(stderr, "error: p = realloc(p, l + 1 + strlen(entry->d_name) + 1)\n");
	exit(1);
      }
      strcpy(p, pathname);
      strcat(p, "/");
      strcat(p, entry->d_name);
      if (!follow_symlinks) {
	struct stat sb;
	if (lstat(p, &sb) == 0 && S_ISLNK(sb.st_mode))
	  continue;
      }
      directory_tree(p, follow_symlinks);
    }
    closedir(d);
    free(p);
    free(entry);
  }
  else {
    switch (errno) {
    case ENOTDIR:	break;
    default:
      fprintf(stderr, "error: %s: \"%s\"\n", strerror(errno), pathname);
      break;
    }
  }
  return 0;
}
int usage(int argc, char *argv[], FILE *fp)
{
  (void)argc;
  fprintf(fp, "%s [-h|--help] [-L] pathname [...]\n", argv[0]);
  return (fp == stderr) ? 1 : 0;
}
int main(int argc, char *argv[])
{
  int rv = 0, a;
  int follow_symlinks = 0;
  if (argc < 2) return usage(argc, argv, stderr);
  for (a=1; a<argc; a++)
    if (strcmp(argv[a], "-h") == 0 || strcmp(argv[a], "--help") == 0)
      return usage(argc, argv, stdout);
    else if (strcmp(argv[a], "-L") == 0)
      follow_symlinks = !0;
    else
      rv = directory_tree(argv[a], follow_symlinks);
  return rv;
}
