/*
  samenamesicase.cc - search case-insensitive same named files recursively

  Copyright (C) 2013 Taiji Yamada <taiji@aihara.co.jp>
  This program is distributed under the BSD 2-clause license.
  See http://opensource.org/licenses/BSD-2-Clause
*/
#include <iostream>
#include <vector>
#include <boost/algorithm/string.hpp>	// to_upper(), etc
#include <boost/filesystem.hpp>

namespace fs = boost::filesystem;

inline
bool filename_ignore_case_less(const fs::path &a, const fs::path &b)
{
  std::string A(a.filename().string()), B(b.filename().string());
  boost::to_upper(A);
  boost::to_upper(B);
  return A < B;
}
inline
bool filename_ignore_case_equal_to(const fs::path &a, const fs::path &b)
{
  return !filename_ignore_case_less(a, b) &&
    !filename_ignore_case_less(b, a);
}

int directory_tree(const fs::path &p, bool follow_symlinks)
{
  try {
    if (!fs::exists(p)) return 1;
    if (fs::is_directory(p)) {
      std::vector<fs::path> v;
      copy(fs::directory_iterator(p), fs::directory_iterator(),
	   back_inserter(v));
      std::sort(v.begin(), v.end(), filename_ignore_case_less);

      std::vector<fs::path>::const_iterator it, jt, kt = v.end();
      for (it=v.begin(), jt=it+1; jt!=v.end(); ++it, ++jt) {
	if (filename_ignore_case_equal_to(*it, *jt)) {
	  if (kt!=it) std::cout << it->string() << std::endl;
	  std::cout << (kt=jt)->string() << std::endl;
	}
	else if (kt != v.end()) {
	  std::cout << std::endl;
	  kt = v.end();
	}
      }
      if (kt != v.end())
	std::cout << std::endl;
      for (it=v.begin(); it!=v.end(); ++it) {
	if (!follow_symlinks && is_symlink(*it))
	  continue;
	directory_tree(*it, follow_symlinks);
      }
    }
  }
  catch (const fs::filesystem_error &ex) {
    std::cerr << ex.what() << std::endl;
    return 1;
  }
  return 0;
}

int usage(int argc, char *argv[], std::ostream &os)
{
  (void)argc;
  os << argv[0] << " [-h|--help] [-L] pathname [...]" << std::endl;
  return os == std::cerr ? 1 : 0;
}

int main(int argc, char *argv[])
{
  int rv = 0, a;
  bool follow_symlinks = false;
  if (argc < 2) return usage(argc, argv, std::cerr);
  for (a=1; a<argc; ++a)
    if (std::string(argv[a]) == "--help")
      return usage(argc, argv, std::cout);
    else if (std::string(argv[a]) == "-L")
      follow_symlinks = true;
    else
      rv = directory_tree(fs::path(argv[a]), follow_symlinks);
  return rv;
}
